/*************************************************************************
Copyright (C) 1999, 2000 Philip R. Wilson - <pwilson@drew.edu>

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
***************************************************************************/

/*
 * mod_auth_nds.c: NetWare Directory Services authentication v2.0.1 beta
 * 
 * Guenter Knauf <eflash@gmx.net> - http://www.gknw.de/development/apache/
 *
 * based on mod_auth_nds.c for Apache 1.3 written by:
 * Philip R. Wilson <pwilson@drew.edu> - http://www.users.drew.edu/~pwilson
 * Contributors:
 * Clifton Woolley <jwoolley@wlu.edu>
 *
 * apache1 code removal, apr 1.2.2 and apache 2.2.0 fixes by:
 * Oden Eriksson <oeriksson@mandriva.com>
 *
 *
 */

/* module configuration hints
MODULE-DEFINITION-START
Name: nds_auth_module
ConfigStart
    LIBS="$LIBS -lncp"
    echo " + using -lncp for ncp support"
ConfigEnd
MODULE-DEFINITION-END
*/

#include <ncp/ncplib.h>
#include <ncp/ndslib.h>
#include <ncp/nwcalls.h>
#include "httpd.h"
#include "http_config.h"
#include "http_core.h"
#include "http_log.h"
#include "http_protocol.h"
#include "http_request.h"
#include "apr_strings.h"

#define log_debug(msg) ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_DEBUG, \
                                    0, r->server, msg)

module AP_MODULE_DECLARE_DATA nds_auth_module;

typedef struct {
    int auth_authoritative;           /* on/off */
    int require_pass;                 /* on/off */
    int context_override;             /* on/off */
    int require_ssl;                  /* on/off */
    char *auth_userfile;              /* file w/ list of permitted users */
    char *auth_expired_uri;           /* uri to call if password expired */
    apr_array_header_t *auth_servers; /* netware server(s) to authenticate to */
    apr_array_header_t *contexts;     /* for contextless login */
    char *auth_failmessage;           /* authorization failure message */
    struct ncp_conn *conn;
} nds_auth_config_rec;


static void nds_auth_init(server_rec *s, apr_pool_t *p) {
}

static void *nds_auth_crdconf(apr_pool_t * p, char *d)
{
    nds_auth_config_rec *sec = apr_pcalloc(p, sizeof(nds_auth_config_rec));
    sec->auth_authoritative = 1;    /* let's be authoritative by default */
    sec->require_pass  = 0;         /* allow empty passwords */
    sec->require_ssl = 1;           /* require SSL connection */
    sec->context_override   = 1;    /* include parent's search list by default*/
    sec->auth_expired_uri = NULL;
    sec->auth_failmessage = NULL;
    sec->auth_servers = apr_array_make(p, 3, sizeof(char *));
    sec->contexts = apr_array_make(p, 1, sizeof(char *));
    sec->conn = NULL;

    return sec;
}


static void *nds_auth_dir_merge(apr_pool_t *p, void *parentv, void *childv)
{
    nds_auth_config_rec *parent = (nds_auth_config_rec *) parentv;
    nds_auth_config_rec *child  = (nds_auth_config_rec *) childv;
    nds_auth_config_rec *new    = (nds_auth_config_rec *)apr_pcalloc
                                  (p, sizeof(nds_auth_config_rec));

    new->auth_authoritative = child->auth_authoritative;
    new->auth_userfile      = child->auth_userfile;
    new->auth_failmessage   = child->auth_failmessage;
    new->context_override   = child->context_override;
    new->require_pass       = child->require_pass;
    new->require_ssl        = child->require_ssl;
    new->conn               = child->conn;

    new->auth_servers       = (child->auth_servers->nelts == 0)
                                ? parent->auth_servers
                                : child->auth_servers;
    new->contexts           = (child->context_override)
                                ? child->contexts
                                : (child->contexts == NULL)
                                ? parent->contexts
                                : apr_array_append(p,parent->contexts,
                                                    child->contexts);

    new->auth_expired_uri = (child->auth_expired_uri == NULL)
                            ? parent->auth_expired_uri
                            : child->auth_expired_uri;

    return (void *)new;
}

static int is_valid_user(request_rec *r, char *user, char *userfile)
{
    ap_configfile_t *f;
    char line[MAX_STRING_LEN];
    char *p;
    char *shortuname;

    if (!(ap_pcfg_openfile(&f, r->pool, userfile))) {
        ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
                      "Couldn't open user list file: %s", userfile);
        return 0;
    }

    while (!ap_cfg_getline(line, MAX_STRING_LEN, f)) {
        if ((line[0] == '#' || (!line[0])))
            continue;
        if ((line[0] != '.') && (*user == '.')) {
            p = apr_pstrdup(r->pool, user);
            ++p;
            shortuname = ap_getword_nc(r->pool, &p, '.');

            if (strcasecmp(shortuname, line) == 0) {
                ap_cfg_closefile(f);
                return 1;
            } else { continue; }
        }
        if (strcasecmp(user, line) == 0) {
            ap_cfg_closefile(f);
            return 1;
        }
    }

    ap_cfg_closefile(f);
    return 0;
}


static const char *nds_cache_set_timeout(cmd_parms *cmd, nds_auth_config_rec
                                         *struct_ptr, char *arg)
{
    return "AuthNDSCacheTimeout: You must compile w/ EAPI+MM to use the cache!";
}


static const char *nds_cache_set_unique_cns(cmd_parms *cmd, nds_auth_config_rec
                                            *struct_ptr, char *arg)
{
    return "AuthNDSUniqueCNs: You must compile w/ EAPI+MM to use the cache!";
}


static const char *add_server(cmd_parms *cmd, nds_auth_config_rec *struct_ptr, char *arg)
{
    char **ptr = (char **)apr_array_push(struct_ptr->auth_servers);
    *ptr = apr_pstrdup(cmd->pool, arg);
    return NULL;
}


static const char *add_search_context(cmd_parms *cmd, nds_auth_config_rec *struct_ptr, char *arg)
{
    char **ptr = (char **)apr_array_push(struct_ptr->contexts);
    *ptr = apr_pstrdup(cmd->pool, arg);
    return NULL;
}


static void cleanup_handler(void *data)
{
    request_rec *r = (request_rec *)data;
    nds_auth_config_rec *sec =
        (nds_auth_config_rec *) ap_get_module_config(r->per_dir_config, &nds_auth_module);

    if (sec->conn != NULL)  {
        ncp_close(sec->conn);
        sec->conn = NULL;
    }
}


static int nds_auth_basic(request_rec *r)
{
    nds_auth_config_rec *sec =
    (nds_auth_config_rec *) ap_get_module_config(r->per_dir_config,
                         &nds_auth_module);
    struct sockaddr_ipx wserv_addr;
    struct ncp_conn_spec spec;
    size_t wsa_len;
    long err;
    int not_wserv;
    unsigned short user_u[256];
    const char *sent_pw;
    u_int32_t user_id;
    char **servers;
    char **contexts;
    int i;
    char *uname;
    char server[NW_MAX_SERVER_NAME_LEN];
    char *ssl_on = NULL;

//    ap_register_cleanup(r->pool, (void *)r, cleanup_handler, ap_null_cleanup);

    if ((r->main != NULL) || (r->prev != NULL) || (r->next != NULL))
        return OK;

    /* Check if we have a custom failmessage */
    if (sec->require_ssl && sec->auth_failmessage)
        ap_custom_response(r, HTTP_FORBIDDEN, sec->auth_failmessage);

    ssl_on = (char*)apr_table_get (r->subprocess_env, "HTTPS");
    /* Check if SSL is required */
    if ((sec->require_ssl) && !ssl_on) {
        ap_log_rerror(APLOG_MARK, APLOG_NOERRNO | APLOG_ERR, 0, r,"SSL connection required");
        return HTTP_FORBIDDEN;
    }

    /* Get the password */
    if ((err = ap_get_basic_auth_pw(r, &sent_pw)))
        return err;

    /* if the user hasn't provided a FDN, don't bother */
    if ((*r->user != '.') && (sec->contexts->nelts == 0)) {
        if (!sec->auth_authoritative)
            return DECLINED;
        else {
            ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, 0, r,
                "%s: not a valid FDN: %s", r->user, r->uri);
            ap_note_basic_auth_failure(r);
            return HTTP_UNAUTHORIZED;
        }
    }

    if ((sec->require_pass) && !sent_pw) {
        if (!sec->auth_authoritative)
            return DECLINED;
        ap_log_rerror(APLOG_MARK, APLOG_NOERRNO | APLOG_ERR, 0, r,
            "User %s did not provide a password: %s", r->user,  r->uri);
        ap_note_basic_auth_failure(r);
        return HTTP_UNAUTHORIZED;
    }

    if (strlen(sent_pw) > 255) {
        if (!sec->auth_authoritative)
            return DECLINED;
        ap_log_rerror(APLOG_MARK, APLOG_NOERRNO | APLOG_ERR, 0, r,
            "User %s sent an overly long password: %s", r->user,  r->uri);
        ap_note_basic_auth_failure(r);
        return HTTP_UNAUTHORIZED;
    }

    /* has a server been specified or do we need to find one? */
    if (sec->auth_servers->nelts != 0) {
        servers = (char **)sec->auth_servers->elts;
        for (i = 0; i < sec->auth_servers->nelts; i++) {
            if (strlen(servers[i]) >= NCP_BINDERY_NAME_LEN) {
                ap_log_rerror(APLOG_MARK, APLOG_NOERRNO | APLOG_ERR, 0, r, "Server name %s too long: %s", servers[i], r->uri);
                return HTTP_INTERNAL_SERVER_ERROR;
            }
            strcpy(spec.server, servers[i]);
            spec.uid = ~0; /* make sure that ncp_open() will use a temp. conn */
            spec.user[0] = 0;

            if ((sec->conn = ncp_open(&spec, &err)) == NULL) {
                ap_log_rerror(APLOG_MARK, APLOG_NOERRNO | APLOG_ERR, 0, r,
                    "ncp_open failed: error connecting to server %s: %s",
                    spec.server, r->uri);
                if (i < (sec->auth_servers->nelts - 1))
                    continue;
                else  {
                    ap_log_rerror(APLOG_MARK, APLOG_NOERRNO | APLOG_ERR, 0, r,
                        "Error: Could not connect to listed server(s): %s",
                        r->uri);
                    return HTTP_INTERNAL_SERVER_ERROR;
                }

            } else {
                break;
            }
        }
    } else {
        if ((sec->conn = ncp_open(NULL, &err)) == NULL) {
            ap_log_rerror(APLOG_MARK, APLOG_NOERRNO | APLOG_ERR, 0, r,
              "ncp_open failed: error finding/connecting to a server: %s", r->uri);
            return HTTP_INTERNAL_SERVER_ERROR;
        }
    }

    err = NWCCGetConnInfo(sec->conn, NWCC_INFO_SERVER_NAME, sizeof(server),
            server);

    if (err) {
        ap_log_rerror(APLOG_MARK, APLOG_NOERRNO | APLOG_ERR, 0, r,
            "NWCCGetConnInfo() failed!: %s", r->uri);
        return HTTP_INTERNAL_SERVER_ERROR;
    }

    ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_DEBUG, 0, r,
        "Connected to server: %s", server);

    i = 0;
    do {
        contexts = (char **)sec->contexts->elts;
        uname = (*r->user == '.')
                ? r->user
                : (*contexts[i] == '.')
                    ? apr_pstrcat(r->pool, ".", r->user, contexts[i], NULL)
                    : apr_pstrcat(r->pool, ".", r->user, ".", contexts[i], NULL);
        ++uname;

        if (strlen(uname) >= sizeof(user_u) / sizeof(unsigned short)) {
            err = ERR_NO_SUCH_ENTRY;
            continue;
        }

        strcpy_cu(user_u, uname);
        wsa_len = sizeof(wserv_addr);

        ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_DEBUG, 0, r,
                     "Trying to resolve username %s for %s", uname, r->user);
        err = nds_resolve_name(sec->conn, 0x64, user_u, &user_id, &not_wserv,
                                (struct sockaddr *)&wserv_addr, &wsa_len);
        --uname;

        if (!err) {
            ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_DEBUG, 0, r,
                "Found FDN %s for %s", uname, r->user);
            r->user = uname;
            break;
        }
        else
            err = ERR_NO_SUCH_ENTRY;

        ++i;
    } while (i < sec->contexts->nelts);

    if (err == ERR_NO_SUCH_ENTRY) {
        if (!(sec->auth_authoritative))
            return DECLINED;
        ap_log_rerror(APLOG_MARK, APLOG_NOERRNO | APLOG_ERR, 0, r,
                     "user %s not found in NDS: %s", r->user, r->uri);
        ap_note_basic_auth_failure(r);
        return HTTP_UNAUTHORIZED;
    }

    if (!nds_get_tree_name(sec->conn, NULL, 0)) {
        if ((err = nds_login_auth(sec->conn, r->user + 1, sent_pw))) {
            switch (err) {
            case (ERR_FAILED_AUTHENTICATION):
                ap_log_rerror(APLOG_MARK, APLOG_NOERRNO | APLOG_ERR, 0, r, "user %s password mismatch: %s", r->user, r->uri);
                break;
            case (NWE_LOGIN_MAX_EXCEEDED):
                ap_log_rerror(APLOG_MARK, APLOG_NOERRNO | APLOG_ERR, 0, r, "user %s has too many connections: %s", r->user, r->uri);
                break;
            case (NWE_LOGIN_UNAUTHORIZED_TIME):
                ap_log_rerror(APLOG_MARK, APLOG_NOERRNO | APLOG_ERR, 0, r, "user %s not allowed to login at this time: %s", r->user, r->uri);
                break;
            case (NWE_LOGIN_UNAUTHORIZED_STATION):
                ap_log_rerror(APLOG_MARK, APLOG_NOERRNO | APLOG_ERR, 0, r, "user %s restricted from this address: %s", r->user, r->uri);
                break;
            case (NWE_ACCT_DISABLED):
                ap_log_rerror(APLOG_MARK, APLOG_NOERRNO | APLOG_ERR, 0, r, "user %s account disabled: %s", r->user, r->uri);
                break;
            case (NWE_PASSWORD_INVALID):
                ap_log_rerror(APLOG_MARK, APLOG_NOERRNO | APLOG_ERR, 0, r, "user %s password expired & no grace logins: %s", r->user, r->uri);
                break;
            case (NWE_PASSWORD_EXPIRED):
                ap_log_rerror(APLOG_MARK, APLOG_NOERRNO | APLOG_ERR, 0, r, "user %s password expired: %s", r->user, r->uri);
                if (sec->auth_expired_uri) {
                    r->filename = apr_pstrdup(r->pool, sec->auth_expired_uri);
                    r->handler = "redirect-handler";
                }
                return OK;
                break;
            default:
                ap_log_rerror(APLOG_MARK, APLOG_NOERRNO | APLOG_ERR, 0, r, "Auth failed for %s - Error code 0x%lx: %s", r->user, err, r->uri);
                break;
            }

            ap_note_basic_auth_failure(r);
            return HTTP_UNAUTHORIZED;
        }

        apr_table_set(r->subprocess_env, "AUTH_USERDN", r->user);
        log_debug("Authentication successful");
        return OK;

    } else {
        log_debug("nds_get_tree_name() failed.");
        return HTTP_INTERNAL_SERVER_ERROR;
    }

}


static int nds_check_access(request_rec * r)
{
    nds_auth_config_rec *sec =
      (nds_auth_config_rec *) ap_get_module_config(r->per_dir_config, &nds_auth_module);
    const apr_array_header_t *reqs_arr = ap_requires(r);
    const char *t, *w;
    int method_restricted = 0;
    int m = r->method_number;
    require_line *reqs;
    int x;
    char *uname;
    char *context;

    /* lacking a "requires" directive, any user will do */
    if (!reqs_arr)
        return (OK);

    if (sec->auth_userfile) {
        if (!is_valid_user(r, r->user, sec->auth_userfile)) {
            if (!(sec->auth_authoritative))
                return DECLINED;
            ap_log_rerror(APLOG_MARK, APLOG_NOERRNO | APLOG_ERR, 0, r,
                          "user %s not found in userlist: %s", r->user, r->uri);
            ap_note_basic_auth_failure(r);
            return HTTP_UNAUTHORIZED;
        }
    }

    reqs = (require_line *) reqs_arr->elts;

    for (x = 0; x < reqs_arr->nelts; x++) {
        if (!(reqs[x].method_mask & (1 << m)))
            continue;

        method_restricted = 1;
        t = reqs[x].requirement;
        w = ap_getword_white(r->pool, &t);

        if (strcmp(w, "valid-user") == 0)
            return OK;
        if (strcmp(w, "user") == 0) {
            while (t[0]) {
                w = ap_getword_conf(r->pool, &t);
                if (strcmp(r->user, w) == 0)
                    return OK;
            }
        }
        else if (strcmp(w, "context") == 0) {
            while (t[0]) {
                w = ap_getword_conf(r->pool, &t);
                uname = r->user + 1; /* we now have a leading '.' */
                context = (char *)w;
                /*
                 * a leading '/' on the require argument indicates that the
                 * user is allowed to be in any subcontext of the given context
                 */
                if (*context == '/') {
                    ++context;
                    if (*context == '.')
                        ++context;
                    while (strlen(uname) > (strlen(context) + 1))
                        ++uname;
                    if (*uname != '.')
                        break;
                } else {
                    if (*context == '.')
                        ++context;
                    while (*uname != '.')
                        ++uname;
                }
                ++uname;

                if (strcasecmp(uname, context) == 0)
                    return OK;
            }
        }
        /* FIXME: group support would be nice */
        else if (sec->auth_authoritative) {
            ap_log_rerror(APLOG_MARK, APLOG_NOERRNO | APLOG_ERR, 0, r,
                  "access to %s failed, reason: unknown require"
                  "directive: \"%s\"", r->uri, reqs[x].requirement);
        }
    }

    if (!method_restricted)
        return OK;

    if (!(sec->auth_authoritative))
        return DECLINED;

    ap_log_rerror(APLOG_MARK, APLOG_NOERRNO | APLOG_ERR, 0, r,
           "access to %s failed, reason: user %s not allowed access",
            r->uri, r->user);

    ap_note_basic_auth_failure(r);
    return HTTP_UNAUTHORIZED;
}


static int handler_redirect(request_rec *r)
{
    nds_auth_config_rec *sec =
    (nds_auth_config_rec *) ap_get_module_config(r->per_dir_config, &nds_auth_module);

    ap_internal_redirect(r->filename, r);

    return OK;
}


static const command_rec nds_auth_cmds[] =
{
    AP_INIT_TAKE1("AuthNDSUserFile", ap_set_file_slot,
     (void *) APR_OFFSETOF(nds_auth_config_rec, auth_userfile),
     OR_AUTHCFG, "text file containing acceptable usernames"),
    AP_INIT_TAKE12("AuthNDSServer", add_server, NULL, 
     OR_AUTHCFG, "List of NetWare servers to authenticate to"),
    AP_INIT_FLAG("AuthNDSAuthoritative", ap_set_flag_slot,
     (void *) APR_OFFSETOF(nds_auth_config_rec, auth_authoritative),
     OR_AUTHCFG, "If set to 'Off', access control will be passed to lower modules if the UserID is unknown to this module"),
    AP_INIT_FLAG("AuthNDSRequirePW", ap_set_flag_slot,
     (void *) APR_OFFSETOF(nds_auth_config_rec, require_pass),
     OR_AUTHCFG, "If set to 'On', users with blank passwords cannot authenticate."),
    AP_INIT_FLAG("AuthNDSRequireSSL", ap_set_flag_slot,
     (void *) APR_OFFSETOF(nds_auth_config_rec, require_ssl),
     OR_AUTHCFG, "If set to 'On', SSL connection to the browser must be established."),
    AP_INIT_TAKE1("AuthNDSExpiredURI", ap_set_string_slot,
     (void *) APR_OFFSETOF(nds_auth_config_rec, auth_expired_uri),
     OR_AUTHCFG, "If set, users with expired passwords will be redirected to this URI"),
    AP_INIT_TAKE12("AuthNDSContext", add_search_context, NULL, 
     OR_AUTHCFG, "Adds search context(s) for contextless logins"),
    AP_INIT_FLAG("AuthNDSContextOverride", ap_set_flag_slot,
     (void *) APR_OFFSETOF(nds_auth_config_rec, context_override),
     OR_AUTHCFG, "If set to 'On', parent's search contexts will be ignored." ),
    AP_INIT_TAKE1("AuthNDSCacheTimeout", nds_cache_set_timeout, NULL,
      RSRC_CONF, "Timeout (in seconds) for NDS user->password cache"),
    AP_INIT_FLAG("AuthNDSUniqueCNs", nds_cache_set_unique_cns, NULL,
     RSRC_CONF, "If set, CN->FDN results will be cached"),
    AP_INIT_TAKE1("AuthNDSFailMessage", ap_set_string_slot,
     (void *) APR_OFFSETOF(nds_auth_config_rec, auth_failmessage),
     OR_AUTHCFG, "text message or redirect URL when authorization failure occurs"),
    {NULL}
};

static int auth_nds_init_handler(apr_pool_t *p, apr_pool_t *plog,
                                apr_pool_t *ptemp, server_rec *s)
{
    ap_add_version_component(p, "mod_auth_nds/2.0");
    return OK;
}
        
static void nds_auth_register_hooks(apr_pool_t *p)
{
//    ap_hook_post_read_request(nds_auth_init, NULL, NULL, APR_HOOK_FIRST);
    ap_hook_post_config(auth_nds_init_handler, NULL, NULL, APR_HOOK_MIDDLE);
    ap_hook_check_user_id(nds_auth_basic, NULL, NULL, APR_HOOK_MIDDLE);
    ap_hook_auth_checker(nds_check_access, NULL, NULL, APR_HOOK_MIDDLE);
}


module AP_MODULE_DECLARE_DATA nds_auth_module = {

    STANDARD20_MODULE_STUFF,
    nds_auth_crdconf,        /* Create Directory configuration record */
    nds_auth_dir_merge,      /* Merge  Directory configuration record */
    NULL,                    /* Create Server    configuration record */
    NULL,                    /* Merge  Server    configuration record */
    nds_auth_cmds,           /* Module command table */
    nds_auth_register_hooks  /* Module hook registration(s) */
};
